# Generated by Django 4.2.1 on 2023-05-11 16:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("scheduler", "0012_alter_cronjob_name_alter_repeatablejob_name_and_more")
    ]

    operations = [
        migrations.AlterField(
            model_name="cronjob",
            name="queue",
            field=models.CharField(
                choices=[("default", "default"), ("low", "low"), ("high", "high")],
                help_text="Queue name",
                max_length=255,
                verbose_name="queue",
            ),
        ),
        migrations.AlterField(
            model_name="repeatablejob",
            name="queue",
            field=models.CharField(
                choices=[("default", "default"), ("low", "low"), ("high", "high")],
                help_text="Queue name",
                max_length=255,
                verbose_name="queue",
            ),
        ),
        migrations.AlterField(
            model_name="scheduledjob",
            name="queue",
            field=models.CharField(
                choices=[("default", "default"), ("low", "low"), ("high", "high")],
                help_text="Queue name",
                max_length=255,
                verbose_name="queue",
            ),
        ),
    ]



# Generated by Django 4.1.5 on 2023-01-20 22:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0006_auto_20230118_1640'),
    ]

    operations = [
        migrations.AddField(
            model_name='cronjob',
            name='result_ttl',
            field=models.IntegerField(blank=True,
                                      help_text='The TTL value (in seconds) of the job result. -1: Result never expires, you should delete jobs manually. 0: Result gets deleted immediately. >0: Result expires after n seconds.',
                                      null=True, verbose_name='result ttl'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='queue',
            field=models.CharField(help_text='Queue name', max_length=16, verbose_name='queue'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='queue',
            field=models.CharField(help_text='Queue name', max_length=16, verbose_name='queue'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='queue',
            field=models.CharField(help_text='Queue name', max_length=16, verbose_name='queue'),
        ),
    ]



import os

from django.conf import settings

from scheduler.settings import conf_settings

settings.SCHEDULER_QUEUES = {
    "default": {"HOST": "localhost", "PORT": 6379, "DB": 0, "DEFAULT_TIMEOUT": 500},
    "test": {
        "HOST": "localhost",
        "PORT": 1,
        "DB": 1,
    },
    "sentinel": {
        "SENTINELS": [("localhost", 26736), ("localhost", 26737)],
        "MASTER_NAME": "testmaster",
        "DB": 1,
        "USERNAME": "redis-user",
        "PASSWORD": "secret",
        "SOCKET_TIMEOUT": 10,
        "SENTINEL_KWARGS": {},
    },
    "test1": {
        "HOST": "localhost",
        "PORT": 1,
        "DB": 1,
        "DEFAULT_TIMEOUT": 400,
    },
    "test2": {
        "HOST": "localhost",
        "PORT": 1,
        "DB": 1,
    },
    "test3": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 1,
    },
    "async": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 1,
        "ASYNC": False,
    },
    "url": {
        "URL": "redis://username:password@host:1234/",
        "DB": 4,
    },
    "url_with_db": {
        "URL": "redis://username:password@host:1234/5",
    },
    "url_default_db": {
        "URL": "redis://username:password@host:1234",
    },
    "django_tasks_scheduler_test": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 0,
    },
    "scheduler_scheduler_active_test": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 0,
        "ASYNC": False,
    },
    "scheduler_scheduler_inactive_test": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 0,
        "ASYNC": False,
    },
    "worker_scheduler_active_test": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 0,
        "ASYNC": False,
    },
    "worker_scheduler_inactive_test": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 0,
        "ASYNC": False,
    },
    "django_tasks_scheduler_test2": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 0,
    },
    "test_scheduler": {
        "HOST": "localhost",
        "PORT": 6379,
        "DB": 0,
        "DEFAULT_TIMEOUT": 400,
    },
}
settings.SCHEDULER_CONFIG = dict(
    FAKEREDIS=(os.getenv("FAKEREDIS", "False") == "True"),
)
conf_settings()



# Generated by Django 4.1.7 on 2023-02-19 22:12

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0007_add_result_ttl'),
    ]

    operations = [
        migrations.RenameField(
            model_name='jobarg',
            old_name='str_val',
            new_name='val',
        ),
        migrations.RenameField(
            model_name='jobkwarg',
            old_name='str_val',
            new_name='val',
        ),
        migrations.RemoveField(
            model_name='jobarg',
            name='bool_val',
        ),
        migrations.RemoveField(
            model_name='jobarg',
            name='datetime_val',
        ),
        migrations.RemoveField(
            model_name='jobarg',
            name='int_val',
        ),
        migrations.RemoveField(
            model_name='jobkwarg',
            name='bool_val',
        ),
        migrations.RemoveField(
            model_name='jobkwarg',
            name='datetime_val',
        ),
        migrations.RemoveField(
            model_name='jobkwarg',
            name='int_val',
        ),
        migrations.AlterField(
            model_name='jobarg',
            name='arg_type',
            field=models.CharField(choices=[('str_val', 'string'), ('int_val', 'int'), ('bool_val', 'boolean'),
                                            ('datetime_val', 'Datetime'), ('callable', 'Callable')], default='str_val',
                                   max_length=12, verbose_name='Argument Type'),
        ),
        migrations.AlterField(
            model_name='jobkwarg',
            name='arg_type',
            field=models.CharField(choices=[('str_val', 'string'), ('int_val', 'int'), ('bool_val', 'boolean'),
                                            ('datetime_val', 'Datetime'), ('callable', 'Callable')], default='str_val',
                                   max_length=12, verbose_name='Argument Type'),
        ),
    ]



from html import escape
from math import ceil
from typing import Tuple, Optional

import redis
from django.contrib import admin, messages
from django.contrib.admin.views.decorators import staff_member_required
from django.core.paginator import Paginator
from django.http import JsonResponse
from django.http.response import HttpResponseNotFound, Http404, HttpResponseBadRequest
from django.shortcuts import redirect
from django.shortcuts import render
from django.urls import reverse, resolve
from django.views.decorators.cache import never_cache
from redis.exceptions import ResponseError

from .queues import get_all_workers, get_connection, QueueNotFoundError
from .queues import get_queue as get_queue_base
from .rq_classes import JobExecution, DjangoWorker, DjangoQueue, InvalidJobOperation
from .settings import SCHEDULER_CONFIG, logger


def get_queue(queue_name: str) -> DjangoQueue:
    try:
        return get_queue_base(queue_name)
    except QueueNotFoundError as e:
        logger.error(e)
        raise Http404(e)


def get_worker_executions(worker):
    res = list()
    for queue in worker.queues:
        curr_jobs = queue.get_all_jobs()
        curr_jobs = [j for j in curr_jobs if j.worker_name == worker.name]
        res.extend(curr_jobs)
    return res


# Create your views here.
@never_cache
@staff_member_required
def stats(request):
    context_data = {**admin.site.each_context(request), **get_statistics(run_maintenance_tasks=True)}
    return render(request, "admin/scheduler/stats.html", context_data)


def stats_json(request):
    auth_token = request.headers.get("Authorization")
    token_validation_func = SCHEDULER_CONFIG.TOKEN_VALIDATION_METHOD
    if request.user.is_staff or (token_validation_func and auth_token and token_validation_func(auth_token)):
        return JsonResponse(get_statistics())

    return HttpResponseNotFound()


def get_statistics(run_maintenance_tasks=False):
    from scheduler.settings import QUEUES

    queues = []
    if run_maintenance_tasks:
        workers = get_all_workers()
        for worker in workers:
            worker.clean_registries()
    for queue_name in QUEUES:
        try:
            queue = get_queue(queue_name)
            connection = get_connection(QUEUES[queue_name])
            connection_kwargs = connection.connection_pool.connection_kwargs

            if run_maintenance_tasks:
                queue.clean_registries()

            # Raw access to the first item from left of the redis list.
            # This might not be accurate since new job can be added from the left
            # with `at_front` parameters.
            # Ideally rq should supports Queue.oldest_job

            last_job_id = queue.last_job_id()
            last_job = queue.fetch_job(last_job_id.decode("utf-8")) if last_job_id else None
            if last_job and last_job.enqueued_at:
                oldest_job_timestamp = last_job.enqueued_at.strftime("%Y-%m-%d, %H:%M:%S")
            else:
                oldest_job_timestamp = "-"

            # parse_class and connection_pool are not needed and not JSON serializable
            connection_kwargs.pop("parser_class", None)
            connection_kwargs.pop("connection_pool", None)

            queue_data = dict(
                name=queue.name,
                jobs=queue.count,
                oldest_job_timestamp=oldest_job_timestamp,
                connection_kwargs=connection_kwargs,
                scheduler_pid=queue.scheduler_pid,
                workers=DjangoWorker.count(queue=queue),
                finished_jobs=len(queue.finished_job_registry),
                started_jobs=len(queue.started_job_registry),
                deferred_jobs=len(queue.deferred_job_registry),
                failed_jobs=len(queue.failed_job_registry),
                scheduled_jobs=len(queue.scheduled_job_registry),
                canceled_jobs=len(queue.canceled_job_registry),
            )
            queues.append(queue_data)
        except redis.ConnectionError as e:
            logger.error(f"Could not connect for queue {queue_name}: {e}")
            continue

    return {"queues": queues}


def _get_registry_job_list(queue, registry, page):
    items_per_page = SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE
    num_jobs = len(registry)
    job_list = []

    if num_jobs == 0:
        return job_list, num_jobs, []

    last_page = int(ceil(num_jobs / items_per_page))
    page_range = range(1, last_page + 1)
    offset = items_per_page * (page - 1)
    job_ids = registry.get_job_ids(offset, offset + items_per_page - 1)
    job_list = JobExecution.fetch_many(job_ids, connection=queue.connection)
    remove_job_ids = [job_id for i, job_id in enumerate(job_ids) if job_list[i] is None]
    valid_jobs = [job for job in job_list if job is not None]
    if registry is not queue:
        for job_id in remove_job_ids:
            registry.remove(job_id)

    return valid_jobs, num_jobs, page_range


@never_cache
@staff_member_required
def jobs_view(request, queue_name: str, registry_name: str):
    queue = get_queue(queue_name)
    registry = queue.get_registry(registry_name)
    if registry is None:
        return HttpResponseNotFound()
    title = registry_name.capitalize()
    page = int(request.GET.get("page", 1))
    job_list, num_jobs, page_range = _get_registry_job_list(queue, registry, page)

    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "registry_name": registry_name,
        "registry": registry,
        "jobs": job_list,
        "num_jobs": num_jobs,
        "page": page,
        "page_range": page_range,
        "job_status": title,
    }
    return render(request, "admin/scheduler/jobs.html", context_data)


@never_cache
@staff_member_required
def queue_workers(request, queue_name):
    queue = get_queue(queue_name)
    all_workers = DjangoWorker.all(queue.connection)
    for w in all_workers:
        w.clean_registries()
    worker_list = [worker for worker in all_workers if queue.name in worker.queue_names()]

    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "workers": worker_list,
    }
    return render(request, "admin/scheduler/queue_workers.html", context_data)


@never_cache
@staff_member_required
def workers(request):
    all_workers = get_all_workers()
    worker_list = [worker for worker in all_workers]

    context_data = {
        **admin.site.each_context(request),
        "workers": worker_list,
    }
    return render(request, "admin/scheduler/workers.html", context_data)


@never_cache
@staff_member_required
def worker_details(request, name):
    queue, worker = None, None
    workers = get_all_workers()
    worker = next((w for w in workers if w.name == name), None)

    if worker is None:
        raise Http404(f"Couldn't find worker with this ID: {name}")
    # Convert microseconds to milliseconds
    worker.total_working_time = worker.total_working_time / 1000

    execution_list = get_worker_executions(worker)
    paginator = Paginator(execution_list, SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE)
    page_number = request.GET.get("p", 1)
    page_obj = paginator.get_page(page_number)
    page_range = paginator.get_elided_page_range(page_obj.number)
    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "worker": worker,
        "queue_names": ", ".join(worker.queue_names()),
        "job": worker.get_current_job(),
        "total_working_time": worker.total_working_time * 1000,
        "executions": page_obj,
        "page_range": page_range,
        "page_var": "p",
    }
    return render(request, "admin/scheduler/worker_details.html", context_data)


def _find_job(job_id: str) -> Tuple[Optional[DjangoQueue], Optional[JobExecution]]:
    from scheduler.settings import QUEUES

    for queue_name in QUEUES:
        try:
            queue = get_queue(queue_name)
            job = JobExecution.fetch(job_id, connection=queue.connection)
            if job.origin == queue_name:
                return queue, job
        except Exception:
            pass
    return None, None


@never_cache
@staff_member_required
def job_detail(request, job_id: str):
    queue, job = _find_job(job_id)
    if job is None:
        return HttpResponseBadRequest(f"Job {escape(job_id)} does not exist, maybe its TTL has passed")
    try:
        job.func_name
        data_is_valid = True
    except Exception:
        data_is_valid = False

    try:
        exc_info = job._exc_info
    except AttributeError:
        exc_info = None

    context_data = {
        **admin.site.each_context(request),
        "job": job,
        "dependency_id": job._dependency_id,
        "queue": queue,
        "data_is_valid": data_is_valid,
        "exc_info": exc_info,
    }
    return render(request, "admin/scheduler/job_detail.html", context_data)


@never_cache
@staff_member_required
def clear_queue_registry(request, queue_name, registry_name):
    queue = get_queue(queue_name)
    registry = queue.get_registry(registry_name)
    if registry is None:
        return HttpResponseNotFound()

    next_url = request.META.get("HTTP_REFERER") or reverse("queue_registry_jobs", args=[queue_name, registry_name])
    if request.method == "POST":
        try:
            if registry is queue:
                queue.empty()
            else:
                job_ids = registry.get_job_ids()
                for job_id in job_ids:
                    registry.remove(job_id, delete_job=True)
            messages.info(request, f"You have successfully cleared the {registry_name} jobs in queue {queue.name}")
        except ResponseError as e:
            messages.error(
                request,
                f"error: {e}",
            )
            raise e
        return redirect("queue_registry_jobs", queue_name, registry_name)
    job_ids = registry.get_job_ids()
    job_list = JobExecution.fetch_many(job_ids, connection=queue.connection)
    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "total_jobs": len(registry),
        "action": "empty",
        "jobs": job_list,
        "next_url": next_url,
        "action_url": reverse(
            "queue_clear",
            args=[
                queue_name,
                registry_name,
            ],
        ),
    }
    return render(request, "admin/scheduler/confirm_action.html", context_data)


@never_cache
@staff_member_required
def requeue_all(request, queue_name, registry_name):
    queue = get_queue(queue_name)
    registry = queue.get_registry(registry_name)
    if registry is None:
        return HttpResponseNotFound()
    next_url = request.META.get("HTTP_REFERER") or reverse("queue_registry_jobs", args=[queue_name, registry_name])
    job_ids = registry.get_job_ids()
    if request.method == "POST":
        count = 0
        # Confirmation received
        jobs = JobExecution.fetch_many(job_ids, connection=queue.connection)
        for job in jobs:
            if job is None:
                continue
            try:
                job.requeue()
                count += 1
            except Exception:
                pass

        messages.info(request, f"You have successfully re-queued {count} jobs!")
        return redirect("queue_registry_jobs", queue_name, registry_name)

    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "total_jobs": len(queue.failed_job_registry),
        "action": "requeue",
        "jobs": [queue.fetch_job(job_id) for job_id in job_ids],
        "next_url": next_url,
        "action_url": reverse("queue_requeue_all", args=[queue_name, registry_name]),
    }

    return render(request, "admin/scheduler/confirm_action.html", context_data)


@never_cache
@staff_member_required
def confirm_action(request, queue_name):
    queue = get_queue(queue_name)
    next_url = request.META.get("HTTP_REFERER") or reverse("queue_registry_jobs", args=[queue_name, "queued"])
    try:
        resolve(next_url)
    except Exception:
        messages.warning(request, "Bad followup URL")
        next_url = reverse("queue_registry_jobs", args=[queue_name, "queued"])

    if request.method == "POST" and request.POST.get("action", False):
        # confirm action
        if request.POST.get("_selected_action", False):
            job_id_list = request.POST.getlist("_selected_action")
            context_data = {
                **admin.site.each_context(request),
                "action": request.POST["action"],
                "jobs": [queue.fetch_job(job_id) for job_id in job_id_list],
                "total_jobs": len(job_id_list),
                "queue": queue,
                "next_url": next_url,
                "action_url": reverse(
                    "queue_actions",
                    args=[
                        queue_name,
                    ],
                ),
            }
            return render(request, "admin/scheduler/confirm_action.html", context_data)

    return redirect(next_url)


@never_cache
@staff_member_required
def actions(request, queue_name):
    queue = get_queue(queue_name)
    next_url = request.POST.get("next_url") or reverse("queue_registry_jobs", args=[queue_name, "queued"])
    try:
        resolve(next_url)
    except Exception:
        messages.warning(request, "Bad followup URL")
        next_url = reverse("queue_registry_jobs", args=[queue_name, "queued"])

    action = request.POST.get("action", False)
    job_ids = request.POST.get("job_ids", False)
    if request.method != "POST" or not action or not job_ids:
        return redirect(next_url)
    job_ids = request.POST.getlist("job_ids")
    if action == "delete":
        jobs = JobExecution.fetch_many(job_ids, connection=queue.connection)
        for job in jobs:
            if job is None:
                continue
            # Remove job id from queue and delete the actual job
            queue.remove_job_id(job.id)
            job.delete()
        messages.info(request, f"You have successfully deleted {len(job_ids)} jobs!")
    elif action == "requeue":
        jobs = JobExecution.fetch_many(job_ids, connection=queue.connection)
        for job in jobs:
            if job is None:
                continue
            job.requeue()
        messages.info(request, f"You have successfully re-queued {len(job_ids)}  jobs!")
    elif action == "stop":
        cancelled_jobs = 0
        jobs = JobExecution.fetch_many(job_ids, connection=queue.connection)
        for job in jobs:
            if job is None:
                continue
            try:
                job.stop_execution(queue.connection)
                job.cancel()
                cancelled_jobs += 1
            except Exception as e:
                logger.warning(f"Could not stop job: {e}")
                pass
        messages.info(request, f"You have successfully stopped {cancelled_jobs}  jobs!")
    return redirect(next_url)


SUPPORTED_JOB_ACTIONS = {"requeue", "delete", "enqueue", "cancel"}


@never_cache
@staff_member_required
def job_action(request, job_id: str, action: str):
    queue, job = _find_job(job_id)
    if job is None:
        return HttpResponseBadRequest(f"Job {escape(job_id)} does not exist, maybe its TTL has passed")
    if action not in SUPPORTED_JOB_ACTIONS:
        return HttpResponseNotFound()

    if request.method != "POST":
        context_data = {
            **admin.site.each_context(request),
            "job": job,
            "queue": queue,
            "action": action,
        }
        return render(request, "admin/scheduler/single_job_action.html", context_data)

    try:
        if action == "requeue":
            job.requeue()
            messages.info(request, f"You have successfully re-queued {job.id}")
            return redirect("job_details", job_id)
        elif action == "delete":
            # Remove job id from queue and delete the actual job
            queue.remove_job_id(job.id)
            job.delete()
            messages.info(request, "You have successfully deleted %s" % job.id)
            return redirect("queue_registry_jobs", queue.name, "queued")
        elif action == "enqueue":
            job.delete(remove_from_queue=False)
            queue._enqueue_job(job)
            messages.info(request, "You have successfully enqueued %s" % job.id)
            return redirect("job_details", job_id)
        elif action == "cancel":
            job.cancel()
            messages.info(request, "You have successfully enqueued %s" % job.id)
            return redirect("job_details", job_id)
    except InvalidJobOperation as e:
        logger.warning(f"Could not perform action: {e}")
        messages.warning(request, f"Could not perform action: {e}")
    return redirect("job_details", job_id)



from datetime import datetime
from typing import Callable

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from scheduler import tools

ARG_TYPE_TYPES_DICT = {
    "str": str,
    "int": int,
    "bool": bool,
    "datetime": datetime,
    "callable": Callable,
}


class BaseTaskArg(models.Model):
    class ArgType(models.TextChoices):
        STR = "str", _("string")
        INT = "int", _("int")
        BOOL = "bool", _("boolean")
        DATETIME = "datetime", _("datetime")
        CALLABLE = "callable", _("callable")

    arg_type = models.CharField(
        _("Argument Type"),
        max_length=12,
        choices=ArgType.choices,
        default=ArgType.STR,
    )
    val = models.CharField(_("Argument Value"), blank=True, max_length=255)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()

    def clean(self):
        if self.arg_type not in ARG_TYPE_TYPES_DICT:
            raise ValidationError(
                {
                    "arg_type": ValidationError(
                        _(f"Could not parse {self.arg_type}, options are: {ARG_TYPE_TYPES_DICT.keys()}"), code="invalid"
                    )
                }
            )
        try:
            if self.arg_type == "callable":
                tools.callable_func(self.val)
            elif self.arg_type == "datetime":
                datetime.fromisoformat(self.val)
            elif self.arg_type == "bool":
                if self.val.lower() not in {"true", "false"}:
                    raise ValidationError
            elif self.arg_type == "int":
                int(self.val)
        except Exception:
            raise ValidationError(
                {"arg_type": ValidationError(_(f"Could not parse {self.val} as {self.arg_type}"), code="invalid")}
            )

    def save(self, **kwargs):
        super(BaseTaskArg, self).save(**kwargs)
        self.content_object.save()

    def delete(self, **kwargs):
        super(BaseTaskArg, self).delete(**kwargs)
        self.content_object.save()

    def value(self):
        if self.arg_type == "callable":
            res = tools.callable_func(self.val)()
        elif self.arg_type == "datetime":
            res = datetime.fromisoformat(self.val)
        elif self.arg_type == "bool":
            res = self.val.lower() == "true"
        else:
            res = ARG_TYPE_TYPES_DICT[self.arg_type](self.val)
        return res

    class Meta:
        abstract = True
        ordering = ["id"]


class TaskArg(BaseTaskArg):
    def __str__(self):
        return f"TaskArg[arg_type={self.arg_type},value={self.value()}]"


class TaskKwarg(BaseTaskArg):
    key = models.CharField(max_length=255)

    def __str__(self):
        key, value = self.value()
        return f"TaskKwarg[key={key},arg_type={self.arg_type},value={self.val}]"

    def value(self):
        return self.key, super(TaskKwarg, self).value()



# Generated by Django 4.1.7 on 2023-04-03 00:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0010_queue'),
    ]

    operations = [
        migrations.CreateModel(
            name='Worker',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'verbose_name_plural': ' Workers',
                'permissions': [['view', 'Access admin page']],
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.AlterModelOptions(
            name='queue',
            options={'default_permissions': (), 'managed': False, 'permissions': [['view', 'Access admin page']],
                     'verbose_name_plural': ' Queues'},
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='at_front',
            field=models.BooleanField(blank=True, default=False,
                                      help_text='When queuing the job, add it in the front of the queue', null=True,
                                      verbose_name='At front'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='cron_string',
            field=models.CharField(
                help_text='Define the schedule in a crontab like syntax.\n            Times are in UTC. Use <a href="https://crontab.guru/">crontab.guru</a> to create a cron string.',
                max_length=64, verbose_name='cron string'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='enabled',
            field=models.BooleanField(default=True,
                                      help_text='Should job be scheduled? This field is useful to keep past jobs that should no longer be scheduled',
                                      verbose_name='enabled'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='job_id',
            field=models.CharField(blank=True, editable=False, help_text='Current job_id on queue', max_length=128,
                                   null=True, verbose_name='job id'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='queue',
            field=models.CharField(choices=[('default', 'default'), ('low', 'low'), ('high', 'high')],
                                   help_text='Queue name', max_length=16, verbose_name='queue'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='repeat',
            field=models.PositiveIntegerField(blank=True,
                                              help_text='Number of times to run the job. Leaving this blank means it will run forever.',
                                              null=True, verbose_name='repeat'),
        ),
        migrations.AlterField(
            model_name='cronjob',
            name='result_ttl',
            field=models.IntegerField(blank=True,
                                      help_text='The TTL value (in seconds) of the job result.<br/>\n               -1: Result never expires, you should delete jobs manually. <br/>\n                0: Result gets deleted immediately. <br/>\n                >0: Result expires after n seconds.',
                                      null=True, verbose_name='result ttl'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='at_front',
            field=models.BooleanField(blank=True, default=False,
                                      help_text='When queuing the job, add it in the front of the queue', null=True,
                                      verbose_name='At front'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='enabled',
            field=models.BooleanField(default=True,
                                      help_text='Should job be scheduled? This field is useful to keep past jobs that should no longer be scheduled',
                                      verbose_name='enabled'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='job_id',
            field=models.CharField(blank=True, editable=False, help_text='Current job_id on queue', max_length=128,
                                   null=True, verbose_name='job id'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='queue',
            field=models.CharField(choices=[('default', 'default'), ('low', 'low'), ('high', 'high')],
                                   help_text='Queue name', max_length=16, verbose_name='queue'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='repeat',
            field=models.PositiveIntegerField(blank=True,
                                              help_text='Number of times to run the job. Leaving this blank means it will run forever.',
                                              null=True, verbose_name='repeat'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='result_ttl',
            field=models.IntegerField(blank=True,
                                      help_text='The TTL value (in seconds) of the job result.<br/>\n               -1: Result never expires, you should delete jobs manually. <br/>\n                0: Result gets deleted immediately. <br/>\n                >0: Result expires after n seconds.',
                                      null=True, verbose_name='result ttl'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='at_front',
            field=models.BooleanField(blank=True, default=False,
                                      help_text='When queuing the job, add it in the front of the queue', null=True,
                                      verbose_name='At front'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='enabled',
            field=models.BooleanField(default=True,
                                      help_text='Should job be scheduled? This field is useful to keep past jobs that should no longer be scheduled',
                                      verbose_name='enabled'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='job_id',
            field=models.CharField(blank=True, editable=False, help_text='Current job_id on queue', max_length=128,
                                   null=True, verbose_name='job id'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='queue',
            field=models.CharField(choices=[('default', 'default'), ('low', 'low'), ('high', 'high')],
                                   help_text='Queue name', max_length=16, verbose_name='queue'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='result_ttl',
            field=models.IntegerField(blank=True,
                                      help_text='The TTL value (in seconds) of the job result.<br/>\n               -1: Result never expires, you should delete jobs manually. <br/>\n                0: Result gets deleted immediately. <br/>\n                >0: Result expires after n seconds.',
                                      null=True, verbose_name='result ttl'),
        ),
    ]



# Generated by Django 4.1.5 on 2023-01-13 22:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0004_cronjob_at_front_repeatablejob_at_front_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cronjob',
            name='at_front',
            field=models.BooleanField(blank=True, default=False, null=True, verbose_name='At front'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='at_front',
            field=models.BooleanField(blank=True, default=False, null=True, verbose_name='At front'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='at_front',
            field=models.BooleanField(blank=True, default=False, null=True, verbose_name='At front'),
        ),
    ]



import redis
import valkey
from django.contrib import admin, messages
from django.contrib.contenttypes.admin import GenericStackedInline
from django.utils.translation import gettext_lazy as _

from scheduler import tools
from scheduler.models import CronTask, TaskArg, TaskKwarg, RepeatableTask, ScheduledTask
from scheduler.settings import SCHEDULER_CONFIG, logger
from scheduler.tools import get_job_executions


class HiddenMixin(object):
    class Media:
        js = [
            "admin/js/jquery.init.js",
        ]


class JobArgInline(HiddenMixin, GenericStackedInline):
    model = TaskArg
    extra = 0
    fieldsets = (
        (
            None,
            {
                "fields": (
                    (
                        "arg_type",
                        "val",
                    ),
                ),
            },
        ),
    )


class JobKwargInline(HiddenMixin, GenericStackedInline):
    model = TaskKwarg
    extra = 0
    fieldsets = (
        (
            None,
            {
                "fields": (
                    ("key",),
                    (
                        "arg_type",
                        "val",
                    ),
                ),
            },
        ),
    )


_LIST_DISPLAY_EXTRA = dict(
    CronTask=(
        "cron_string",
        "next_run",
        "successful_runs",
        "last_successful_run",
        "failed_runs",
        "last_failed_run",
    ),
    ScheduledTask=("scheduled_time",),
    RepeatableTask=(
        "scheduled_time",
        "interval_display",
        "successful_runs",
        "last_successful_run",
        "failed_runs",
        "last_failed_run",
    ),
)
_FIELDSET_EXTRA = dict(
    CronTask=(
        "cron_string",
        "timeout",
        "result_ttl",
        (
            "successful_runs",
            "last_successful_run",
        ),
        (
            "failed_runs",
            "last_failed_run",
        ),
    ),
    ScheduledTask=("scheduled_time", "timeout", "result_ttl"),
    RepeatableTask=(
        "scheduled_time",
        (
            "interval",
            "interval_unit",
        ),
        "repeat",
        "timeout",
        "result_ttl",
        (
            "successful_runs",
            "last_successful_run",
        ),
        (
            "failed_runs",
            "last_failed_run",
        ),
    ),
)


@admin.register(CronTask, ScheduledTask, RepeatableTask)
class TaskAdmin(admin.ModelAdmin):
    """TaskAdmin admin view for all task models.
    Using the _LIST_DISPLAY_EXTRA and _FIELDSET_EXTRA additional data for each model.
    """

    save_on_top = True
    change_form_template = "admin/scheduler/change_form.html"
    actions = [
        "disable_selected",
        "enable_selected",
        "enqueue_job_now",
    ]
    inlines = [
        JobArgInline,
        JobKwargInline,
    ]
    list_filter = ("enabled",)
    list_display = (
        "enabled",
        "name",
        "job_id",
        "function_string",
        "is_scheduled",
        "queue",
    )
    list_display_links = ("name",)
    readonly_fields = ("job_id",)
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "callable",
                    "enabled",
                    "at_front",
                ),
            },
        ),
        (
            _("RQ Settings"),
            {
                "fields": (
                    "queue",
                    "job_id",
                ),
            },
        ),
    )

    def get_list_display(self, request):
        if self.model.__name__ not in _LIST_DISPLAY_EXTRA:
            raise ValueError(f"Unrecognized model {self.model}")
        return TaskAdmin.list_display + _LIST_DISPLAY_EXTRA[self.model.__name__]

    def get_fieldsets(self, request, obj=None):
        if self.model.__name__ not in _FIELDSET_EXTRA:
            raise ValueError(f"Unrecognized model {self.model}")
        return TaskAdmin.fieldsets + (
            (
                _("Scheduling"),
                {
                    "fields": _FIELDSET_EXTRA[self.model.__name__],
                },
            ),
        )

    @admin.display(description="Next run")
    def next_run(self, o: CronTask):
        return tools.get_next_cron_time(o.cron_string)

    def change_view(self, request, object_id, form_url="", extra_context=None):
        extra = extra_context or {}
        obj = self.get_object(request, object_id)
        try:
            execution_list = get_job_executions(obj.queue, obj)
        except (redis.ConnectionError, valkey.ConnectionError) as e:
            logger.warn(f"Could not get job executions: {e}")
            execution_list = list()
        paginator = self.get_paginator(request, execution_list, SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE)
        page_number = request.GET.get("p", 1)
        page_obj = paginator.get_page(page_number)
        page_range = paginator.get_elided_page_range(page_obj.number)

        extra.update(
            {
                "pagination_required": paginator.count > SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE,
                "executions": page_obj,
                "page_range": page_range,
                "page_var": "p",
            }
        )

        return super(TaskAdmin, self).change_view(request, object_id, form_url, extra_context=extra)

    def delete_queryset(self, request, queryset):
        for job in queryset:
            job.unschedule()
        super(TaskAdmin, self).delete_queryset(request, queryset)

    def delete_model(self, request, obj):
        obj.unschedule()
        super(TaskAdmin, self).delete_model(request, obj)

    @admin.action(description=_("Disable selected %(verbose_name_plural)s"), permissions=("change",))
    def disable_selected(self, request, queryset):
        rows_updated = 0
        for obj in queryset.filter(enabled=True).iterator():
            obj.enabled = False
            obj.unschedule()
            rows_updated += 1

        message_bit = "1 job was" if rows_updated == 1 else f"{rows_updated} jobs were"

        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(request, f"{message_bit} successfully disabled and unscheduled.", level=level)

    @admin.action(description=_("Enable selected %(verbose_name_plural)s"), permissions=("change",))
    def enable_selected(self, request, queryset):
        rows_updated = 0
        for obj in queryset.filter(enabled=False).iterator():
            obj.enabled = True
            obj.save()
            rows_updated += 1

        message_bit = "1 job was" if rows_updated == 1 else f"{rows_updated} jobs were"
        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(request, f"{message_bit} successfully enabled and scheduled.", level=level)

    @admin.action(description="Enqueue now", permissions=("change",))
    def enqueue_job_now(self, request, queryset):
        task_names = []
        for task in queryset:
            task.enqueue_to_run()
            task_names.append(task.name)
        self.message_user(
            request,
            f"The following jobs have been enqueued: {', '.join(task_names)}",
        )



# Generated by Django 4.2 on 2023-04-18 19:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0011_worker_alter_queue_options_alter_cronjob_at_front_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cronjob',
            name='name',
            field=models.CharField(help_text='Name of the job.', max_length=128, unique=True, verbose_name='name'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='name',
            field=models.CharField(help_text='Name of the job.', max_length=128, unique=True, verbose_name='name'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='name',
            field=models.CharField(help_text='Name of the job.', max_length=128, unique=True, verbose_name='name'),
        ),
    ]



import time

from django.test import TestCase

from scheduler import job, settings
from . import test_settings  # noqa
from ..queues import get_queue, QueueNotFoundError


@job
def test_job():
    time.sleep(1)
    return 1 + 1


@job("django_tasks_scheduler_test")
def test_job_diff_queue():
    time.sleep(1)
    return 1 + 1


@job(timeout=1)
def test_job_timeout():
    time.sleep(1)
    return 1 + 1


@job(result_ttl=1)
def test_job_result_ttl():
    return 1 + 1


class JobDecoratorTest(TestCase):
    def setUp(self) -> None:
        get_queue("default").connection.flushall()

    def test_job_decorator_no_params(self):
        test_job.delay()
        config = settings.SCHEDULER_CONFIG
        self._assert_job_with_func_and_props("default", test_job, config.DEFAULT_RESULT_TTL, config.DEFAULT_TIMEOUT)

    def test_job_decorator_timeout(self):
        test_job_timeout.delay()
        config = settings.SCHEDULER_CONFIG
        self._assert_job_with_func_and_props("default", test_job_timeout, config.DEFAULT_RESULT_TTL, 1)

    def test_job_decorator_result_ttl(self):
        test_job_result_ttl.delay()
        config = settings.SCHEDULER_CONFIG
        self._assert_job_with_func_and_props("default", test_job_result_ttl, 1, config.DEFAULT_TIMEOUT)

    def test_job_decorator_different_queue(self):
        test_job_diff_queue.delay()
        config = settings.SCHEDULER_CONFIG
        self._assert_job_with_func_and_props(
            "django_tasks_scheduler_test",
            test_job_diff_queue,
            config.DEFAULT_RESULT_TTL,
            config.DEFAULT_TIMEOUT,
        )

    def _assert_job_with_func_and_props(self, queue_name, expected_func, expected_result_ttl, expected_timeout):
        queue = get_queue(queue_name)
        jobs = queue.get_jobs()
        self.assertEqual(1, len(jobs))

        j = jobs[0]
        self.assertEqual(j.func, expected_func)
        self.assertEqual(j.result_ttl, expected_result_ttl)
        self.assertEqual(j.timeout, expected_timeout)

    def test_job_decorator_bad_queue(self):
        with self.assertRaises(QueueNotFoundError):

            @job("bad-queue")
            def test_job_bad_queue():
                time.sleep(1)
                return 1 + 1



# Generated by Django 4.0.1 on 2022-01-06 20:40

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0001_initial_squashed_0005_added_result_ttl'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cronjob',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='repeatablejob',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='scheduledjob',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



# Generated by Django 5.1b1 on 2024-06-29 14:21

import scheduler.models.scheduled_task
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('scheduler', '0017_remove_crontask_repeat_crontask_failed_runs_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='crontask',
            name='queue',
            field=models.CharField(choices=scheduler.models.scheduled_task.get_queue_choices, help_text='Queue name', max_length=255, verbose_name='queue'),
        ),
        migrations.AlterField(
            model_name='repeatabletask',
            name='queue',
            field=models.CharField(choices=scheduler.models.scheduled_task.get_queue_choices, help_text='Queue name', max_length=255, verbose_name='queue'),
        ),
        migrations.AlterField(
            model_name='scheduledtask',
            name='queue',
            field=models.CharField(choices=scheduler.models.scheduled_task.get_queue_choices, help_text='Queue name', max_length=255, verbose_name='queue'),
        ),
    ]



import math
import uuid
from datetime import timedelta
from typing import Dict

import croniter
from django.apps import apps
from django.conf import settings as django_settings
from django.contrib import admin
from django.contrib.contenttypes.fields import GenericRelation
from django.core.exceptions import ValidationError
from django.core.mail import mail_admins
from django.db import models
from django.templatetags.tz import utc
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from scheduler import settings
from scheduler import tools
from scheduler.models.args import TaskArg, TaskKwarg
from scheduler.queues import get_queue
from scheduler.rq_classes import DjangoQueue
from scheduler.settings import QUEUES
from scheduler.settings import logger

SCHEDULER_INTERVAL = settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL


def failure_callback(job, connection, result, *args, **kwargs):
    model_name = job.meta.get("task_type", None)
    if model_name is None:
        return
    model = apps.get_model(app_label="scheduler", model_name=model_name)
    task = model.objects.filter(job_id=job.id).first()
    if task is None:
        logger.warn(f"Could not find {model_name} task for job {job.id}")
        return
    mail_admins(
        f"Task {task.id}/{task.name} has failed",
        "See django-admin for logs",
    )
    task.job_id = None
    if isinstance(task, (CronTask, RepeatableTask)):
        task.failed_runs += 1
        task.last_failed_run = timezone.now()
    task.save(schedule_job=True)


def success_callback(job, connection, result, *args, **kwargs):
    model_name = job.meta.get("task_type", None)
    if model_name is None:
        return
    model = apps.get_model(app_label="scheduler", model_name=model_name)
    task = model.objects.filter(job_id=job.id).first()
    if task is None:
        return
    task.job_id = None
    if isinstance(task, (CronTask, RepeatableTask)):
        task.successful_runs += 1
        task.last_successful_run = timezone.now()
    task.save(schedule_job=True)


def get_queue_choices():
    return [(queue, queue) for queue in QUEUES.keys()]


class BaseTask(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    TASK_TYPE = "BaseTask"
    name = models.CharField(
        _("name"),
        max_length=128,
        unique=True,
        help_text="Name of the job.",
    )
    callable = models.CharField(_("callable"), max_length=2048)
    callable_args = GenericRelation(TaskArg, related_query_name="args")
    callable_kwargs = GenericRelation(TaskKwarg, related_query_name="kwargs")
    enabled = models.BooleanField(
        _("enabled"),
        default=True,
        help_text=_(
            "Should job be scheduled? This field is useful to keep " "past jobs that should no longer be scheduled"
        ),
    )
    queue = models.CharField(
        _("queue"),
        max_length=255,
        choices=get_queue_choices,
        help_text=_("Queue name"),
    )
    job_id = models.CharField(
        _("job id"), max_length=128, editable=False, blank=True, null=True, help_text=_("Current job_id on queue")
    )
    at_front = models.BooleanField(
        _("At front"),
        default=False,
        blank=True,
        null=True,
        help_text=_("When queuing the job, add it in the front of the queue"),
    )
    timeout = models.IntegerField(
        _("timeout"),
        blank=True,
        null=True,
        help_text=_(
            "Timeout specifies the maximum runtime, in seconds, for the job "
            "before it'll be considered 'lost'. Blank uses the default "
            "timeout."
        ),
    )
    result_ttl = models.IntegerField(
        _("result ttl"),
        blank=True,
        null=True,
        help_text=mark_safe(
            """The TTL value (in seconds) of the job result.<br/>
               -1: Result never expires, you should delete jobs manually. <br/>
                0: Result gets deleted immediately. <br/>
                >0: Result expires after n seconds."""
        ),
    )

    def callable_func(self):
        """Translate callable string to callable"""
        return tools.callable_func(self.callable)

    @admin.display(boolean=True, description=_("is scheduled?"))
    def is_scheduled(self) -> bool:
        """Check whether a next job for this task is queued/scheduled to be executed"""
        if self.job_id is None:  # no job_id => is not scheduled
            return False
        # check whether job_id is in scheduled/queued/active jobs
        scheduled_jobs = self.rqueue.scheduled_job_registry.get_job_ids()
        enqueued_jobs = self.rqueue.get_job_ids()
        active_jobs = self.rqueue.started_job_registry.get_job_ids()
        res = (self.job_id in scheduled_jobs) or (self.job_id in enqueued_jobs) or (self.job_id in active_jobs)
        # If the job_id is not scheduled/queued/started,
        # update the job_id to None. (The job_id belongs to a previous run which is completed)
        if not res:
            self.job_id = None
            super(BaseTask, self).save()
        return res

    @admin.display(description="Callable")
    def function_string(self) -> str:
        args = self.parse_args()
        args_list = [repr(arg) for arg in args]
        kwargs = self.parse_kwargs()
        kwargs_list = [k + "=" + repr(v) for (k, v) in kwargs.items()]
        return self.callable + f"({', '.join(args_list + kwargs_list)})"

    def parse_args(self):
        """Parse args for running the job"""
        args = self.callable_args.all()
        return [arg.value() for arg in args]

    def parse_kwargs(self):
        """Parse kwargs for running the job"""
        kwargs = self.callable_kwargs.all()
        return dict([kwarg.value() for kwarg in kwargs])

    def _next_job_id(self):
        addition = uuid.uuid4().hex[-10:]
        name = self.name.replace("/", ".")
        return f"{self.queue}:{name}:{addition}"

    def _enqueue_args(self) -> Dict:
        """Args for DjangoQueue.enqueue.
        Set all arguments for DjangoQueue.enqueue/enqueue_at.
        Particularly:
        - set job timeout and ttl
        - ensure a callback to reschedule the job next iteration.
        - Set job-id to proper format
        - set job meta
        """
        res = dict(
            meta=dict(
                task_type=self.TASK_TYPE,
                scheduled_task_id=self.id,
            ),
            on_success=success_callback,
            on_failure=failure_callback,
            job_id=self._next_job_id(),
        )
        if self.at_front:
            res["at_front"] = self.at_front
        if self.timeout:
            res["job_timeout"] = self.timeout
        if self.result_ttl is not None:
            res["result_ttl"] = self.result_ttl
        return res

    @property
    def rqueue(self) -> DjangoQueue:
        """Returns django-queue for job"""
        return get_queue(self.queue)

    def ready_for_schedule(self) -> bool:
        """Is the task ready to be scheduled?

        If the task is already scheduled or disabled, then it is not
        ready to be scheduled.

        :returns: True if the task is ready to be scheduled.
        """
        if self.is_scheduled():
            logger.debug(f"Task {self.name} already scheduled")
            return False
        if not self.enabled:
            logger.debug(f"Task {str(self)} disabled, enable task before scheduling")
            return False
        return True

    def schedule(self) -> bool:
        """Schedule the next execution for the task to run.
        :returns: True if a job was scheduled, False otherwise.
        """
        if not self.ready_for_schedule():
            return False
        schedule_time = self._schedule_time()
        kwargs = self._enqueue_args()
        job = self.rqueue.enqueue_at(
            schedule_time,
            tools.run_task,
            args=(self.TASK_TYPE, self.id),
            **kwargs,
        )
        self.job_id = job.id
        super(BaseTask, self).save()
        return True

    def enqueue_to_run(self) -> bool:
        """Enqueue job to run now."""
        kwargs = self._enqueue_args()
        job = self.rqueue.enqueue(
            tools.run_task,
            args=(self.TASK_TYPE, self.id),
            **kwargs,
        )
        self.job_id = job.id
        self.save(schedule_job=False)
        return True

    def unschedule(self) -> bool:
        """Remove a job from django-queue.

        If a job is queued to be executed or scheduled to be executed, it will remove it.
        """
        queue = self.rqueue
        if self.job_id is None:
            return True
        queue.remove(self.job_id)
        queue.scheduled_job_registry.remove(self.job_id)
        self.job_id = None
        self.save(schedule_job=False)
        return True

    def _schedule_time(self):
        return utc(self.scheduled_time) if django_settings.USE_TZ else self.scheduled_time

    def to_dict(self) -> Dict:
        """Export model to dictionary, so it can be saved as external file backup"""
        res = dict(
            model=self.TASK_TYPE,
            name=self.name,
            callable=self.callable,
            callable_args=[
                dict(
                    arg_type=arg.arg_type,
                    val=arg.val,
                )
                for arg in self.callable_args.all()
            ],
            callable_kwargs=[
                dict(
                    arg_type=arg.arg_type,
                    key=arg.key,
                    val=arg.val,
                )
                for arg in self.callable_kwargs.all()
            ],
            enabled=self.enabled,
            queue=self.queue,
            repeat=getattr(self, "repeat", None),
            at_front=self.at_front,
            timeout=self.timeout,
            result_ttl=self.result_ttl,
            cron_string=getattr(self, "cron_string", None),
            scheduled_time=self._schedule_time().isoformat(),
            interval=getattr(self, "interval", None),
            interval_unit=getattr(self, "interval_unit", None),
            successful_runs=getattr(self, "successful_runs", None),
            failed_runs=getattr(self, "failed_runs", None),
            last_successful_run=getattr(self, "last_successful_run", None),
            last_failed_run=getattr(self, "last_failed_run", None),
        )
        return res

    def get_absolute_url(self):
        model = self._meta.model.__name__.lower()
        return reverse(
            f"admin:scheduler_{model}_change",
            args=[
                self.id,
            ],
        )

    def __str__(self):
        func = self.function_string()
        return f"{self.TASK_TYPE}[{self.name}={func}]"

    def save(self, **kwargs):
        schedule_job = kwargs.pop("schedule_job", True)
        update_fields = kwargs.get("update_fields", None)
        if update_fields:
            kwargs["update_fields"] = set(update_fields).union({"modified"})
        super(BaseTask, self).save(**kwargs)
        if schedule_job:
            self.schedule()
            super(BaseTask, self).save()

    def delete(self, **kwargs):
        self.unschedule()
        super(BaseTask, self).delete(**kwargs)

    def clean_callable(self):
        try:
            tools.callable_func(self.callable)
        except Exception:
            raise ValidationError(
                {"callable": ValidationError(_("Invalid callable, must be importable"), code="invalid")}
            )

    def clean_queue(self):
        queue_keys = settings.QUEUES.keys()
        if self.queue not in queue_keys:
            raise ValidationError(
                {
                    "queue": ValidationError(
                        _("Invalid queue, must be one of: {}".format(", ".join(queue_keys))), code="invalid"
                    )
                }
            )

    def clean(self):
        self.clean_queue()
        self.clean_callable()

    class Meta:
        abstract = True


class ScheduledTimeMixin(models.Model):
    scheduled_time = models.DateTimeField(_("scheduled time"))

    class Meta:
        abstract = True


class RepeatableMixin(models.Model):
    failed_runs = models.PositiveIntegerField(
        _("failed runs"),
        default=0,
        help_text=_("Number of times the task has failed"),
    )
    successful_runs = models.PositiveIntegerField(
        _("successful runs"),
        default=0,
        help_text=_("Number of times the task has succeeded"),
    )
    last_successful_run = models.DateTimeField(
        _("last successful run"),
        blank=True,
        null=True,
        help_text=_("Last time the task has succeeded"),
    )
    last_failed_run = models.DateTimeField(
        _("last failed run"),
        blank=True,
        null=True,
        help_text=_("Last time the task has failed"),
    )

    class Meta:
        abstract = True


class ScheduledTask(ScheduledTimeMixin, BaseTask):
    TASK_TYPE = "ScheduledTask"

    def ready_for_schedule(self) -> bool:
        return super(ScheduledTask, self).ready_for_schedule() and (
            self.scheduled_time is None or self.scheduled_time >= timezone.now()
        )

    class Meta:
        verbose_name = _("Scheduled Task")
        verbose_name_plural = _("Scheduled Tasks")
        ordering = ("name",)


class RepeatableTask(RepeatableMixin, ScheduledTimeMixin, BaseTask):
    class TimeUnits(models.TextChoices):
        SECONDS = "seconds", _("seconds")
        MINUTES = "minutes", _("minutes")
        HOURS = "hours", _("hours")
        DAYS = "days", _("days")
        WEEKS = "weeks", _("weeks")

    interval = models.PositiveIntegerField(_("interval"))
    interval_unit = models.CharField(
        _("interval unit"), max_length=12, choices=TimeUnits.choices, default=TimeUnits.HOURS
    )
    repeat = models.PositiveIntegerField(
        _("repeat"),
        blank=True,
        null=True,
        help_text=_("Number of times to run the job. Leaving this blank means it will run forever."),
    )
    TASK_TYPE = "RepeatableTask"

    def clean(self):
        super(RepeatableTask, self).clean()
        self.clean_interval_unit()
        self.clean_result_ttl()

    def clean_interval_unit(self):
        if SCHEDULER_INTERVAL > self.interval_seconds():
            raise ValidationError(
                _("Job interval is set lower than %(queue)r queue's interval. " "minimum interval is %(interval)"),
                code="invalid",
                params={"queue": self.queue, "interval": SCHEDULER_INTERVAL},
            )
        if self.interval_seconds() % SCHEDULER_INTERVAL:
            raise ValidationError(
                _("Job interval is not a multiple of rq_scheduler's interval frequency: %(interval)ss"),
                code="invalid",
                params={"interval": SCHEDULER_INTERVAL},
            )

    def clean_result_ttl(self) -> None:
        """
        Throws an error if there are repeats left to run and the result_ttl won't last until the next scheduled time.
        :return: None
        """
        if self.result_ttl and self.result_ttl != -1 and self.result_ttl < self.interval_seconds() and self.repeat:
            raise ValidationError(
                _(
                    "Job result_ttl must be either indefinite (-1) or "
                    "longer than the interval, %(interval)s seconds, to ensure rescheduling."
                ),
                code="invalid",
                params={"interval": self.interval_seconds()},
            )

    def interval_display(self):
        return "{} {}".format(self.interval, self.get_interval_unit_display())

    def interval_seconds(self):
        kwargs = {
            self.interval_unit: self.interval,
        }
        return timedelta(**kwargs).total_seconds()

    def _enqueue_args(self):
        res = super(RepeatableTask, self)._enqueue_args()
        res["meta"]["interval"] = self.interval_seconds()
        res["meta"]["repeat"] = self.repeat
        return res

    def _schedule_time(self):
        _now = timezone.now()
        if self.scheduled_time >= _now:
            return super()._schedule_time()
        gap = math.ceil((_now.timestamp() - self.scheduled_time.timestamp()) / self.interval_seconds())
        if self.repeat is None or self.repeat >= gap:
            self.scheduled_time += timedelta(seconds=self.interval_seconds() * gap)
            self.repeat = (self.repeat - gap) if self.repeat is not None else None
        return super()._schedule_time()

    def ready_for_schedule(self):
        if super(RepeatableTask, self).ready_for_schedule() is False:
            return False
        if self._schedule_time() < timezone.now():
            return False
        return True

    class Meta:
        verbose_name = _("Repeatable Task")
        verbose_name_plural = _("Repeatable Tasks")
        ordering = ("name",)


class CronTask(RepeatableMixin, BaseTask):
    TASK_TYPE = "CronTask"

    cron_string = models.CharField(
        _("cron string"),
        max_length=64,
        help_text=mark_safe(
            """Define the schedule in a crontab like syntax.
            Times are in UTC. Use <a href="https://crontab.guru/">crontab.guru</a> to create a cron string."""
        ),
    )

    def clean(self):
        super(CronTask, self).clean()
        self.clean_cron_string()

    def clean_cron_string(self):
        try:
            croniter.croniter(self.cron_string)
        except ValueError as e:
            raise ValidationError({"cron_string": ValidationError(_(str(e)), code="invalid")})

    def _schedule_time(self):
        self.scheduled_time = tools.get_next_cron_time(self.cron_string)
        return super()._schedule_time()

    class Meta:
        verbose_name = _("Cron Task")
        verbose_name_plural = _("Cron Tasks")
        ordering = ("name",)



from typing import List, Dict

import redis
import valkey

from .connection_types import RedisSentinel, BrokerConnectionClass
from .rq_classes import JobExecution, DjangoQueue, DjangoWorker
from .settings import SCHEDULER_CONFIG
from .settings import logger, Broker

_CONNECTION_PARAMS = {
    "URL",
    "DB",
    "USE_REDIS_CACHE",
    "UNIX_SOCKET_PATH",
    "HOST",
    "PORT",
    "PASSWORD",
    "SENTINELS",
    "MASTER_NAME",
    "SOCKET_TIMEOUT",
    "SSL",
    "CONNECTION_KWARGS",
}


class QueueNotFoundError(Exception):
    pass


def _get_redis_connection(config, use_strict_redis=False):
    """
    Returns a redis connection from a connection config
    """
    if SCHEDULER_CONFIG.BROKER == Broker.FAKEREDIS:
        import fakeredis

        redis_cls = fakeredis.FakeRedis if use_strict_redis else fakeredis.FakeStrictRedis
    else:
        redis_cls = BrokerConnectionClass[(SCHEDULER_CONFIG.BROKER, use_strict_redis)]
    logger.debug(f"Getting connection for {config}")
    if "URL" in config:
        if config.get("SSL") or config.get("URL").startswith("rediss://"):
            return redis_cls.from_url(
                config["URL"],
                db=config.get("DB"),
                ssl_cert_reqs=config.get("SSL_CERT_REQS", "required"),
            )
        else:
            return redis_cls.from_url(
                config["URL"],
                db=config.get("DB"),
            )
    if "UNIX_SOCKET_PATH" in config:
        return redis_cls(unix_socket_path=config["UNIX_SOCKET_PATH"], db=config["DB"])

    if "SENTINELS" in config:
        connection_kwargs = {
            "db": config.get("DB"),
            "password": config.get("PASSWORD"),
            "username": config.get("USERNAME"),
            "socket_timeout": config.get("SOCKET_TIMEOUT"),
        }
        connection_kwargs.update(config.get("CONNECTION_KWARGS", {}))
        sentinel_kwargs = config.get("SENTINEL_KWARGS", {})
        sentinel = RedisSentinel(config["SENTINELS"], sentinel_kwargs=sentinel_kwargs, **connection_kwargs)
        return sentinel.master_for(
            service_name=config["MASTER_NAME"],
            redis_class=redis_cls,
        )

    return redis_cls(
        host=config["HOST"],
        port=config["PORT"],
        db=config.get("DB", 0),
        username=config.get("USERNAME", None),
        password=config.get("PASSWORD"),
        ssl=config.get("SSL", False),
        ssl_cert_reqs=config.get("SSL_CERT_REQS", "required"),
        **config.get("REDIS_CLIENT_KWARGS", {}),
    )


def get_connection(queue_settings, use_strict_redis=False):
    """Returns a Redis connection to use based on parameters in SCHEDULER_QUEUES"""
    return _get_redis_connection(queue_settings, use_strict_redis)


def get_queue(
    name="default", default_timeout=None, is_async=None, autocommit=None, connection=None, **kwargs
) -> DjangoQueue:
    """Returns an DjangoQueue using parameters defined in `SCHEDULER_QUEUES`"""
    from .settings import QUEUES

    if name not in QUEUES:
        raise QueueNotFoundError(f"Queue {name} not found, queues={QUEUES.keys()}")
    queue_settings = QUEUES[name]
    if is_async is None:
        is_async = queue_settings.get("ASYNC", True)

    if default_timeout is None:
        default_timeout = queue_settings.get("DEFAULT_TIMEOUT")
    if connection is None:
        connection = get_connection(queue_settings)
    return DjangoQueue(
        name, default_timeout=default_timeout, connection=connection, is_async=is_async, autocommit=autocommit, **kwargs
    )


def get_all_workers():
    from .settings import QUEUES

    workers = set()
    for queue_name in QUEUES:
        connection = get_connection(QUEUES[queue_name])
        try:
            curr_workers = set(DjangoWorker.all(connection=connection))
            workers.update(curr_workers)
        except (redis.ConnectionError, valkey.ConnectionError) as e:
            logger.error(f"Could not connect for queue {queue_name}: {e}")
    return workers


def _queues_share_connection_params(q1_params: Dict, q2_params: Dict):
    """Check that both queues share the same connection parameters"""
    return all(
        ((p not in q1_params and p not in q2_params) or (q1_params.get(p, None) == q2_params.get(p, None)))
        for p in _CONNECTION_PARAMS
    )


def get_queues(*queue_names, **kwargs) -> List[DjangoQueue]:
    """Return queue instances from specified queue names.
    All instances must use the same Redis connection.
    """
    from .settings import QUEUES

    kwargs["job_class"] = JobExecution
    queue_params = QUEUES[queue_names[0]]
    queues = [get_queue(queue_names[0], **kwargs)]
    # perform consistency checks while building return list
    for name in queue_names[1:]:
        if not _queues_share_connection_params(queue_params, QUEUES[name]):
            raise ValueError(
                f'Queues must have the same redis connection. "{name}" and'
                f' "{queue_names[0]}" have different connections'
            )
        queue = get_queue(name, **kwargs)
        queues.append(queue)

    return queues



# Generated by Django 5.0.1 on 2024-01-10 17:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('scheduler', '0016_rename_jobarg_taskarg_rename_jobkwarg_taskkwarg_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='crontask',
            name='repeat',
        ),
        migrations.AddField(
            model_name='crontask',
            name='failed_runs',
            field=models.PositiveIntegerField(default=0, help_text='Number of times the task has failed', verbose_name='failed runs'),
        ),
        migrations.AddField(
            model_name='crontask',
            name='last_failed_run',
            field=models.DateTimeField(blank=True, help_text='Last time the task has failed', null=True, verbose_name='last failed run'),
        ),
        migrations.AddField(
            model_name='crontask',
            name='last_successful_run',
            field=models.DateTimeField(blank=True, help_text='Last time the task has succeeded', null=True, verbose_name='last successful run'),
        ),
        migrations.AddField(
            model_name='crontask',
            name='successful_runs',
            field=models.PositiveIntegerField(default=0, help_text='Number of times the task has succeeded', verbose_name='successful runs'),
        ),
        migrations.AddField(
            model_name='repeatabletask',
            name='failed_runs',
            field=models.PositiveIntegerField(default=0, help_text='Number of times the task has failed', verbose_name='failed runs'),
        ),
        migrations.AddField(
            model_name='repeatabletask',
            name='last_failed_run',
            field=models.DateTimeField(blank=True, help_text='Last time the task has failed', null=True, verbose_name='last failed run'),
        ),
        migrations.AddField(
            model_name='repeatabletask',
            name='last_successful_run',
            field=models.DateTimeField(blank=True, help_text='Last time the task has succeeded', null=True, verbose_name='last successful run'),
        ),
        migrations.AddField(
            model_name='repeatabletask',
            name='successful_runs',
            field=models.PositiveIntegerField(default=0, help_text='Number of times the task has succeeded', verbose_name='successful runs'),
        ),
    ]



import click
from django.core.management.base import BaseCommand

from scheduler.queues import get_queue
from scheduler.rq_classes import JobExecution


class Command(BaseCommand):
    help = "Delete failed jobs from Django queue."

    def add_arguments(self, parser):
        parser.add_argument("--queue", "-q", dest="queue", default="default", help="Specify the queue [default]")
        parser.add_argument("-f", "--func", help='optional job function name, e.g. "app.tasks.func"')
        parser.add_argument("--dry-run", action="store_true", help="Do not actually delete failed jobs")

    def handle(self, *args, **options):
        queue = get_queue(options.get("queue", "default"))
        job_ids = queue.failed_job_registry.get_job_ids()
        jobs = JobExecution.fetch_many(job_ids, connection=queue.connection)
        func_name = options.get("func", None)
        if func_name is not None:
            jobs = [job for job in jobs if job.func_name == func_name]
        dry_run = options.get("dry_run", False)
        click.echo(f"Found {len(jobs)} failed jobs")
        for job in jobs:
            click.echo(f"Deleting {job.id}")
            if not dry_run:
                job.delete()
        click.echo(f"Deleted {len(jobs)} failed jobs")



from django.db import models


class Worker(models.Model):
    """Placeholder model with no database table, but with django admin page
    and contenttype permission"""

    class Meta:
        managed = False  # not in Django's database
        default_permissions = ()
        permissions = [["view", "Access admin page"]]
        verbose_name_plural = " Workers"



from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from scheduler.models import TaskArg, TaskKwarg
from .jobs import arg_callable
from .testtools import taskarg_factory


class TestAllTaskArg(TestCase):
    TaskArgClass = TaskArg

    def test_bad_arg_type(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="bad_arg_type", val="something")
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_one_value_invalid_str_int(self):
        arg = taskarg_factory(
            self.TaskArgClass,
            arg_type="int",
            val="not blank",
        )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_callable_invalid(self):
        arg = taskarg_factory(
            self.TaskArgClass,
            arg_type="callable",
            val="bad_callable",
        )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_datetime_invalid(self):
        arg = taskarg_factory(
            self.TaskArgClass,
            arg_type="datetime",
            val="bad datetime",
        )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_bool_invalid(self):
        arg = taskarg_factory(
            self.TaskArgClass,
            arg_type="bool",
            val="bad bool",
        )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_int_invalid(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="int", val="str")
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_str_clean(self):
        arg = taskarg_factory(self.TaskArgClass, val="something")
        self.assertIsNone(arg.clean())


class TestTaskArg(TestCase):
    TaskArgClass = TaskArg

    def test_str(self):
        arg = taskarg_factory(self.TaskArgClass)
        self.assertEqual(f"TaskArg[arg_type={arg.arg_type},value={arg.value()}]", str(arg))

    def test_value(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="str", val="something")
        self.assertEqual(arg.value(), "something")

    def test__str__str_val(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="str", val="something")
        self.assertEqual("something", str(arg.value()))

    def test__str__int_val(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="int", val="1")
        self.assertEqual("1", str(arg.value()))

    def test__str__datetime_val(self):
        _time = timezone.now()
        arg = taskarg_factory(self.TaskArgClass, arg_type="datetime", val=str(_time))
        self.assertEqual(str(_time), str(arg.value()))

    def test__str__bool_val(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="bool", val="True")
        self.assertEqual("True", str(arg.value()))

    def test__repr__str_val(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="str", val="something")
        self.assertEqual("'something'", repr(arg.value()))

    def test__repr__int_val(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="int", val="1")
        self.assertEqual("1", repr(arg.value()))

    def test__repr__datetime_val(self):
        _time = timezone.now()
        arg = taskarg_factory(self.TaskArgClass, arg_type="datetime", val=str(_time))
        self.assertEqual(repr(_time), repr(arg.value()))

    def test__repr__bool_val(self):
        arg = taskarg_factory(self.TaskArgClass, arg_type="bool", val="False")
        self.assertEqual("False", repr(arg.value()))

    def test_callable_arg_type__clean(self):
        method = arg_callable
        arg = taskarg_factory(
            self.TaskArgClass,
            arg_type="callable",
            val=f"{method.__module__}.{method.__name__}",
        )
        self.assertIsNone(arg.clean())
        self.assertEqual(1, arg.value())
        self.assertEqual(2, arg.value())


class TestTaskKwarg(TestAllTaskArg):
    TaskArgClass = TaskKwarg

    def test_str(self):
        arg = taskarg_factory(self.TaskArgClass)
        self.assertEqual(f"TaskKwarg[key={arg.key},arg_type={arg.arg_type},value={arg.val}]", str(arg))

    def test_value(self):
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="str", val="value")
        self.assertEqual(kwarg.value(), ("key", "value"))

    def test__str__str_val(self):
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="str", val="something")
        self.assertEqual("TaskKwarg[key=key,arg_type=str,value=something]", str(kwarg))

    def test__str__int_val(self):
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="int", val=1)
        self.assertEqual("TaskKwarg[key=key,arg_type=int,value=1]", str(kwarg))

    def test__str__datetime_val(self):
        _time = timezone.now()
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="datetime", val=str(_time))
        self.assertEqual(f"TaskKwarg[key=key,arg_type=datetime,value={_time}]", str(kwarg))

    def test__str__bool_val(self):
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="bool", val="True")
        self.assertEqual("TaskKwarg[key=key,arg_type=bool,value=True]", str(kwarg))

    def test__repr__str_val(self):
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="str", val="something")
        self.assertEqual("('key', 'something')", repr(kwarg.value()))

    def test__repr__int_val(self):
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="int", val="1")
        self.assertEqual("('key', 1)", repr(kwarg.value()))

    def test__repr__datetime_val(self):
        _time = timezone.now()
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="datetime", val=str(_time))
        self.assertEqual("('key', {})".format(repr(_time)), repr(kwarg.value()))

    def test__repr__bool_val(self):
        kwarg = taskarg_factory(self.TaskArgClass, key="key", arg_type="bool", val="True")
        self.assertEqual("('key', True)", repr(kwarg.value()))



from django.urls import path

from . import views

urlpatterns = [
    path("queues/", views.stats, name="queues_home"),
    path("queues/stats.json", views.stats_json, name="queues_home_json"),
    path("queues/<str:queue_name>/workers/", views.queue_workers, name="queue_workers"),
    path("queues/<str:queue_name>/<str:registry_name>/jobs", views.jobs_view, name="queue_registry_jobs"),
    path("queues/<str:queue_name>/<str:registry_name>/empty/", views.clear_queue_registry, name="queue_clear"),
    path("queues/<str:queue_name>/<str:registry_name>/requeue-all/", views.requeue_all, name="queue_requeue_all"),
    path("queues/<str:queue_name>/confirm-action/", views.confirm_action, name="queue_confirm_action"),
    path("queues/<str:queue_name>/actions/", views.actions, name="queue_actions"),
]

urlpatterns += [
    path("workers/", views.workers, name="workers_home"),
    path("workers/<str:name>/", views.worker_details, name="worker_details"),
    path("jobs/<str:job_id>/", views.job_detail, name="job_details"),
    path("jobs/<str:job_id>/<str:action>/", views.job_action, name="queue_job_action"),
]



import click
from django.core.management.base import BaseCommand

from scheduler.queues import get_queue


class Command(BaseCommand):
    """
    Queues the function given with the first argument with the
    parameters given with the rest of the argument list.
    """

    help = __doc__
    args = "<function arg arg ...>"

    def add_arguments(self, parser):
        parser.add_argument("--queue", "-q", dest="queue", default="default", help="Specify the queue [default]")
        parser.add_argument("--timeout", "-t", type=int, dest="timeout", help="A timeout in seconds")
        parser.add_argument(
            "--result-ttl", "-r", type=int, dest="result_ttl", help="Time to store job results in seconds"
        )
        parser.add_argument(
            "callable",
            help="Method to call",
        )
        parser.add_argument("args", nargs="*", help="Args for callable")

    def handle(self, **options):
        verbosity = int(options.get("verbosity", 1))
        timeout = options.get("timeout")
        result_ttl = options.get("result_ttl")
        queue = get_queue(options.get("queue"))
        func = options.get("callable")
        args = options.get("args")
        job = queue.enqueue_call(func, args=args, timeout=timeout, result_ttl=result_ttl)
        if verbosity:
            click.echo(f"Job {job.id} created")



from scheduler import settings
from .queues import get_queue, QueueNotFoundError
from .rq_classes import rq_job_decorator


def job(*args, **kwargs):
    """
    The same as rq package's job decorator, but it automatically works out
    the ``connection`` argument from SCHEDULER_QUEUES.

    And also, it allows simplified ``@job`` syntax to put a job into the default queue.

    """
    if len(args) == 0:
        func = None
        queue = "default"
    else:
        if callable(args[0]):
            func = args[0]
            queue = "default"
        else:
            func = None
            queue = args[0]
        args = args[1:]

    if isinstance(queue, str):
        try:
            queue = get_queue(queue)
            if "connection" not in kwargs:
                kwargs["connection"] = queue.connection
        except KeyError:
            raise QueueNotFoundError(f"Queue {queue} does not exist")

    kwargs.setdefault("result_ttl", settings.SCHEDULER_CONFIG.DEFAULT_RESULT_TTL)
    kwargs.setdefault("timeout", settings.SCHEDULER_CONFIG.DEFAULT_TIMEOUT)

    decorator = rq_job_decorator(queue, *args, **kwargs)
    if func:
        return decorator(func)
    return decorator



import os
import uuid

from rq.job import Job
from scheduler.rq_classes import JobExecution
from scheduler.tests.testtools import SchedulerBaseCase
from scheduler.tools import create_worker
from . import test_settings  # noqa
from .. import settings


class TestWorker(SchedulerBaseCase):
    def test_create_worker__two_workers_same_queue(self):
        worker1 = create_worker("default", "django_tasks_scheduler_test")
        worker1.register_birth()
        worker2 = create_worker("default")
        worker2.register_birth()
        hostname = os.uname()[1]
        self.assertEqual(f"{hostname}-worker.1", worker1.name)
        self.assertEqual(f"{hostname}-worker.2", worker2.name)

    def test_create_worker__worker_with_queues_different_connection(self):
        with self.assertRaises(ValueError):
            create_worker("default", "test1")

    def test_create_worker__with_name(self):
        name = uuid.uuid4().hex
        worker1 = create_worker("default", name=name)
        self.assertEqual(name, worker1.name)

    def test_create_worker__with_name_containing_slash(self):
        name = uuid.uuid4().hex[-4:] + "/" + uuid.uuid4().hex[-4:]
        worker1 = create_worker("default", name=name)
        self.assertEqual(name.replace("/", "."), worker1.name)

    def test_create_worker__scheduler_interval(self):
        prev = settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL
        settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 1
        worker = create_worker("default")
        worker.work(burst=True)
        self.assertEqual(worker.scheduler.interval, 1)
        settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL = prev

    def test_get_worker_with_custom_job_class(self):
        # Test with string representation of job_class
        worker = create_worker("default", job_class="scheduler.rq_classes.JobExecution")
        self.assertTrue(issubclass(worker.job_class, Job))
        self.assertTrue(issubclass(worker.job_class, JobExecution))

    def test_get_worker_without_custom_job_class(self):
        worker = create_worker("default")
        self.assertTrue(issubclass(worker.job_class, JobExecution))



# Generated by Django 4.1.5 on 2023-01-18 16:40

from django.db import migrations


def forwards_func(apps, schema_editor):
    pass


def reverse_func(apps, schema_editor):
    # forwards_func() creates two Country instances,
    # so reverse_func() should delete them.
    cronjob_model = apps.get_model(app_label='scheduler', model_name='CronJob')
    db_alias = schema_editor.connection.alias
    cronjob_model.objects.using(db_alias).filter(name='Job scheduling jobs').delete()


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0005_alter_cronjob_at_front_alter_repeatablejob_at_front_and_more'),
    ]

    operations = [
        migrations.RunPython(forwards_func, reverse_func),
    ]



import zoneinfo
from datetime import datetime, timedelta

from django.contrib.messages import get_messages
from django.core.exceptions import ValidationError
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from freezegun import freeze_time

from scheduler import settings
from scheduler.models import BaseTask, TaskArg, TaskKwarg, ScheduledTask
from scheduler.tools import run_task, create_worker
from . import jobs
from .testtools import (
    task_factory,
    taskarg_factory,
    _get_job_from_scheduled_registry,
    SchedulerBaseCase,
    _get_executions,
)
from ..queues import get_queue


def assert_response_has_msg(response, message):
    messages = [m.message for m in get_messages(response.wsgi_request)]
    assert message in messages, f'expected "{message}" in {messages}'


def assert_has_execution_with_status(task, status):
    job_list = _get_executions(task)
    job_list = [(j.id, j.get_status()) for j in job_list]
    for job in job_list:
        if job[1] == status:
            return
    raise AssertionError(f"{task} does not have an execution with status {status}: {job_list}")


class BaseTestCases:
    class TestBaseTask(SchedulerBaseCase):
        TaskModelClass = BaseTask

        def test_callable_func(self):
            task = task_factory(self.TaskModelClass)
            task.callable = "scheduler.tests.jobs.test_job"
            func = task.callable_func()
            self.assertEqual(jobs.test_job, func)

        def test_callable_func_not_callable(self):
            task = task_factory(self.TaskModelClass)
            task.callable = "scheduler.tests.jobs.test_non_callable"
            with self.assertRaises(TypeError):
                task.callable_func()

        def test_clean_callable(self):
            task = task_factory(self.TaskModelClass)
            task.callable = "scheduler.tests.jobs.test_job"
            self.assertIsNone(task.clean_callable())

        def test_clean_callable_invalid(self):
            task = task_factory(self.TaskModelClass)
            task.callable = "scheduler.tests.jobs.test_non_callable"
            with self.assertRaises(ValidationError):
                task.clean_callable()

        def test_clean_queue(self):
            for queue in settings.QUEUES.keys():
                task = task_factory(self.TaskModelClass)
                task.queue = queue
                self.assertIsNone(task.clean_queue())

        def test_clean_queue_invalid(self):
            task = task_factory(self.TaskModelClass)
            task.queue = "xxxxxx"
            task.callable = "scheduler.tests.jobs.test_job"
            with self.assertRaises(ValidationError):
                task.clean()

        # next 2 check the above are included in job.clean() function
        def test_clean_base(self):
            task = task_factory(self.TaskModelClass)
            task.queue = list(settings.QUEUES)[0]
            task.callable = "scheduler.tests.jobs.test_job"
            self.assertIsNone(task.clean())

        def test_clean_invalid_callable(self):
            task = task_factory(self.TaskModelClass)
            task.queue = list(settings.QUEUES)[0]
            task.callable = "scheduler.tests.jobs.test_non_callable"
            with self.assertRaises(ValidationError):
                task.clean()

        def test_clean_invalid_queue(self):
            task = task_factory(self.TaskModelClass)
            task.queue = "xxxxxx"
            task.callable = "scheduler.tests.jobs.test_job"
            with self.assertRaises(ValidationError):
                task.clean()

        def test_is_schedulable_already_scheduled(self):
            task = task_factory(
                self.TaskModelClass,
            )
            task.schedule()
            self.assertTrue(task.is_scheduled())

        def test_is_schedulable_disabled(self):
            task = task_factory(self.TaskModelClass)
            task.enabled = False
            self.assertFalse(task.enabled)

        def test_schedule(self):
            task = task_factory(
                self.TaskModelClass,
            )
            self.assertTrue(task.is_scheduled())
            self.assertIsNotNone(task.job_id)

        def test_unschedulable(self):
            task = task_factory(self.TaskModelClass, enabled=False)
            self.assertFalse(task.is_scheduled())
            self.assertIsNone(task.job_id)

        def test_unschedule(self):
            task = task_factory(
                self.TaskModelClass,
            )
            self.assertTrue(task.unschedule())
            self.assertIsNone(task.job_id)

        def test_unschedule_not_scheduled(self):
            task = task_factory(self.TaskModelClass, enabled=False)
            self.assertTrue(task.unschedule())
            self.assertIsNone(task.job_id)

        def test_save_enabled(self):
            task = task_factory(
                self.TaskModelClass,
            )
            self.assertIsNotNone(task.job_id)

        def test_save_disabled(self):
            task = task_factory(self.TaskModelClass, enabled=False)
            task.save()
            self.assertIsNone(task.job_id)

        def test_save_and_schedule(self):
            task = task_factory(
                self.TaskModelClass,
            )
            self.assertIsNotNone(task.job_id)
            self.assertTrue(task.is_scheduled())

        def test_schedule2(self):
            task = task_factory(self.TaskModelClass)
            task.queue = list(settings.QUEUES)[0]
            task.enabled = False
            task.scheduled_time = timezone.now() + timedelta(minutes=1)
            self.assertFalse(task.schedule())

        def test_delete_and_unschedule(self):
            task = task_factory(
                self.TaskModelClass,
            )
            self.assertIsNotNone(task.job_id)
            self.assertTrue(task.is_scheduled())
            task.delete()
            self.assertFalse(task.is_scheduled())

        def test_job_create(self):
            prev_count = self.TaskModelClass.objects.count()
            task_factory(self.TaskModelClass)
            self.assertEqual(self.TaskModelClass.objects.count(), prev_count + 1)

        def test_str(self):
            name = "test"
            task = task_factory(self.TaskModelClass, name=name)
            self.assertEqual(f"{self.TaskModelClass.__name__}[{name}={task.callable}()]", str(task))

        def test_callable_passthrough(self):
            task = task_factory(self.TaskModelClass)
            entry = _get_job_from_scheduled_registry(task)
            self.assertEqual(entry.func, run_task)
            job_model, job_id = entry.args
            self.assertEqual(job_model, self.TaskModelClass.__name__)
            self.assertEqual(job_id, task.id)

        def test_timeout_passthrough(self):
            task = task_factory(self.TaskModelClass, timeout=500)
            entry = _get_job_from_scheduled_registry(task)
            self.assertEqual(entry.timeout, 500)

        def test_at_front_passthrough(self):
            task = task_factory(self.TaskModelClass, at_front=True)
            queue = task.rqueue
            jobs_to_schedule = queue.scheduled_job_registry.get_job_ids()
            self.assertIn(task.job_id, jobs_to_schedule)

        def test_callable_result(self):
            task = task_factory(
                self.TaskModelClass,
            )
            entry = _get_job_from_scheduled_registry(task)
            self.assertEqual(entry.perform(), 2)

        def test_callable_empty_args_and_kwargs(self):
            task = task_factory(self.TaskModelClass, callable="scheduler.tests.jobs.test_args_kwargs")
            entry = _get_job_from_scheduled_registry(task)
            self.assertEqual(entry.perform(), "test_args_kwargs()")

        def test_delete_args(self):
            task = task_factory(
                self.TaskModelClass,
            )
            arg = taskarg_factory(TaskArg, val="one", content_object=task)
            self.assertEqual(1, task.callable_args.count())
            arg.delete()
            self.assertEqual(0, task.callable_args.count())

        def test_delete_kwargs(self):
            task = task_factory(
                self.TaskModelClass,
            )
            kwarg = taskarg_factory(TaskKwarg, key="key1", arg_type="str", val="one", content_object=task)
            self.assertEqual(1, task.callable_kwargs.count())
            kwarg.delete()
            self.assertEqual(0, task.callable_kwargs.count())

        def test_parse_args(self):
            task = task_factory(
                self.TaskModelClass,
            )
            date = timezone.now()
            taskarg_factory(TaskArg, val="one", content_object=task)
            taskarg_factory(TaskArg, arg_type="int", val=2, content_object=task)
            taskarg_factory(TaskArg, arg_type="bool", val=True, content_object=task)
            taskarg_factory(TaskArg, arg_type="bool", val=False, content_object=task)
            taskarg_factory(TaskArg, arg_type="datetime", val=date, content_object=task)
            self.assertEqual(task.parse_args(), ["one", 2, True, False, date])

        def test_parse_kwargs(self):
            job = task_factory(
                self.TaskModelClass,
            )
            date = timezone.now()
            taskarg_factory(TaskKwarg, key="key1", arg_type="str", val="one", content_object=job)
            taskarg_factory(TaskKwarg, key="key2", arg_type="int", val=2, content_object=job)
            taskarg_factory(TaskKwarg, key="key3", arg_type="bool", val=True, content_object=job)
            taskarg_factory(TaskKwarg, key="key4", arg_type="datetime", val=date, content_object=job)
            kwargs = job.parse_kwargs()
            self.assertEqual(kwargs, dict(key1="one", key2=2, key3=True, key4=date))

        def test_callable_args_and_kwargs(self):
            task = task_factory(self.TaskModelClass, callable="scheduler.tests.jobs.test_args_kwargs")
            date = timezone.now()
            taskarg_factory(TaskArg, arg_type="str", val="one", content_object=task)
            taskarg_factory(TaskKwarg, key="key1", arg_type="int", val=2, content_object=task)
            taskarg_factory(TaskKwarg, key="key2", arg_type="datetime", val=date, content_object=task)
            taskarg_factory(TaskKwarg, key="key3", arg_type="bool", val=False, content_object=task)
            task.save()
            entry = _get_job_from_scheduled_registry(task)
            self.assertEqual(entry.perform(), "test_args_kwargs('one', key1=2, key2={}, key3=False)".format(date))

        def test_function_string(self):
            task = task_factory(
                self.TaskModelClass,
            )
            date = timezone.now()
            taskarg_factory(TaskArg, arg_type="str", val="one", content_object=task)
            taskarg_factory(TaskArg, arg_type="int", val="1", content_object=task)
            taskarg_factory(TaskArg, arg_type="datetime", val=date, content_object=task)
            taskarg_factory(TaskArg, arg_type="bool", val=True, content_object=task)
            taskarg_factory(TaskKwarg, key="key1", arg_type="str", val="one", content_object=task)
            taskarg_factory(TaskKwarg, key="key2", arg_type="int", val=2, content_object=task)
            taskarg_factory(TaskKwarg, key="key3", arg_type="datetime", val=date, content_object=task)
            taskarg_factory(TaskKwarg, key="key4", arg_type="bool", val=False, content_object=task)
            self.assertEqual(
                task.function_string(),
                f"scheduler.tests.jobs.test_job('one', 1, {repr(date)}, True, "
                f"key1='one', key2=2, key3={repr(date)}, key4=False)",
            )

        def test_admin_list_view(self):
            # arrange
            self.client.login(username="admin", password="admin")
            job = task_factory(
                self.TaskModelClass,
            )
            model = job._meta.model.__name__.lower()
            url = reverse(f"admin:scheduler_{model}_changelist")
            # act
            res = self.client.get(url)
            # assert
            self.assertEqual(200, res.status_code)

        def test_admin_list_view_delete_model(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(
                self.TaskModelClass,
            )
            model = task._meta.model.__name__.lower()
            url = reverse(f"admin:scheduler_{model}_changelist")
            # act
            res = self.client.post(
                url,
                data={
                    "action": "delete_model",
                    "_selected_action": [
                        task.pk,
                    ],
                },
            )
            # assert
            self.assertEqual(302, res.status_code)

        def test_admin_run_job_now_enqueues_job_at(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(
                self.TaskModelClass,
            )
            model = task._meta.model.__name__.lower()
            url = reverse(f"admin:scheduler_{model}_changelist")
            # act
            res = self.client.post(
                url,
                data={
                    "action": "enqueue_job_now",
                    "_selected_action": [
                        task.pk,
                    ],
                },
            )
            # assert
            self.assertEqual(302, res.status_code)
            task.refresh_from_db()
            queue = get_queue(task.queue)
            self.assertIn(task.job_id, queue.get_job_ids())

        def test_admin_change_view(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(
                self.TaskModelClass,
            )
            model = task._meta.model.__name__.lower()
            url = reverse(
                f"admin:scheduler_{model}_change",
                args=[
                    task.pk,
                ],
            )
            # act
            res = self.client.get(url)
            # assert
            self.assertEqual(200, res.status_code)

        def test_admin_change_view__bad_redis_connection(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(self.TaskModelClass, queue="test2", instance_only=True)
            task.save(schedule_job=False)
            model = task._meta.model.__name__.lower()
            url = reverse(
                f"admin:scheduler_{model}_change",
                args=[
                    task.pk,
                ],
            )
            # act
            res = self.client.get(url)
            # assert
            self.assertEqual(200, res.status_code)

        def test_admin_enqueue_job_now(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(
                self.TaskModelClass,
            )
            self.assertIsNotNone(task.job_id)
            self.assertTrue(task.is_scheduled())
            data = {
                "action": "enqueue_job_now",
                "_selected_action": [
                    task.id,
                ],
            }
            model = task._meta.model.__name__.lower()
            url = reverse(f"admin:scheduler_{model}_changelist")
            # act
            res = self.client.post(url, data=data, follow=True)

            # assert part 1
            self.assertEqual(200, res.status_code)
            entry = _get_job_from_scheduled_registry(task)
            task_model, scheduled_task_id = entry.args
            self.assertEqual(task_model, task.TASK_TYPE)
            self.assertEqual(scheduled_task_id, task.id)
            self.assertEqual("scheduled", entry.get_status())
            assert_has_execution_with_status(task, "queued")

            # act 2
            worker = create_worker(
                "default",
                fork_job_execution=False,
            )
            worker.work(burst=True)

            # assert 2
            entry = _get_job_from_scheduled_registry(task)
            self.assertEqual(task_model, task.TASK_TYPE)
            self.assertEqual(scheduled_task_id, task.id)
            assert_has_execution_with_status(task, "finished")

        def test_admin_enable_job(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(self.TaskModelClass, enabled=False)
            self.assertIsNone(task.job_id)
            self.assertFalse(task.is_scheduled())
            data = {
                "action": "enable_selected",
                "_selected_action": [
                    task.id,
                ],
            }
            model = task._meta.model.__name__.lower()
            url = reverse(f"admin:scheduler_{model}_changelist")
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            task.refresh_from_db()
            self.assertTrue(task.enabled)
            self.assertTrue(task.is_scheduled())
            assert_response_has_msg(res, "1 job was successfully enabled and scheduled.")

        def test_admin_disable_job(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(self.TaskModelClass, enabled=True)
            task.save()
            data = {
                "action": "disable_selected",
                "_selected_action": [
                    task.id,
                ],
            }
            model = task._meta.model.__name__.lower()
            url = reverse(f"admin:scheduler_{model}_changelist")
            self.assertTrue(task.is_scheduled())
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            task.refresh_from_db()
            self.assertFalse(task.is_scheduled())
            self.assertFalse(task.enabled)
            assert_response_has_msg(res, "1 job was successfully disabled and unscheduled.")

        def test_admin_single_delete(self):
            # arrange
            self.client.login(username="admin", password="admin")
            prev_count = self.TaskModelClass.objects.count()
            task = task_factory(
                self.TaskModelClass,
            )
            self.assertIsNotNone(task.job_id)
            self.assertTrue(task.is_scheduled())
            prev = len(_get_executions(task))
            model = task._meta.model.__name__.lower()
            url = reverse(
                f"admin:scheduler_{model}_delete",
                args=[
                    task.pk,
                ],
            )
            data = {
                "post": "yes",
            }
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            self.assertEqual(prev_count, self.TaskModelClass.objects.count())
            self.assertEqual(prev - 1, len(_get_executions(task)))

        def test_admin_delete_selected(self):
            # arrange
            self.client.login(username="admin", password="admin")
            task = task_factory(self.TaskModelClass, enabled=True)
            task.save()
            queue = get_queue(task.queue)
            scheduled_jobs = queue.scheduled_job_registry.get_job_ids()
            job_id = task.job_id
            self.assertIn(job_id, scheduled_jobs)
            data = {
                "action": "delete_selected",
                "_selected_action": [
                    task.id,
                ],
                "post": "yes",
            }
            model = task._meta.model.__name__.lower()
            url = reverse(f"admin:scheduler_{model}_changelist")
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            assert_response_has_msg(res, f"Successfully deleted 1 {self.TaskModelClass._meta.verbose_name}.")
            self.assertIsNone(self.TaskModelClass.objects.filter(id=task.id).first())
            scheduled_jobs = queue.scheduled_job_registry.get_job_ids()
            self.assertNotIn(job_id, scheduled_jobs)

    class TestSchedulableJob(TestBaseTask):
        # Currently ScheduledJob and RepeatableJob
        TaskModelClass = ScheduledTask

        @freeze_time("2016-12-25")
        @override_settings(USE_TZ=False)
        def test_schedule_time_no_tz(self):
            task = task_factory(self.TaskModelClass)
            task.scheduled_time = datetime(2016, 12, 25, 8, 0, 0, tzinfo=None)
            self.assertEqual("2016-12-25T08:00:00", task._schedule_time().isoformat())

        @freeze_time("2016-12-25")
        @override_settings(USE_TZ=True)
        def test_schedule_time_with_tz(self):
            task = task_factory(self.TaskModelClass)
            est = zoneinfo.ZoneInfo("US/Eastern")
            task.scheduled_time = datetime(2016, 12, 25, 8, 0, 0, tzinfo=est)
            self.assertEqual("2016-12-25T13:00:00+00:00", task._schedule_time().isoformat())

        def test_result_ttl_passthrough(self):
            job = task_factory(self.TaskModelClass, result_ttl=500)
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(entry.result_ttl, 500)


class TestScheduledJob(BaseTestCases.TestSchedulableJob):
    TaskModelClass = ScheduledTask

    def test_clean(self):
        job = task_factory(self.TaskModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = "scheduler.tests.jobs.test_job"
        self.assertIsNone(job.clean())

    def test_unschedulable_old_job(self):
        job = task_factory(self.TaskModelClass, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertFalse(job.is_scheduled())



import logging
import os
import sys

import click
import redis
import valkey
from django.core.management.base import BaseCommand
from django.db import connections
from rq.logutils import setup_loghandlers

from scheduler.rq_classes import register_sentry
from scheduler.tools import create_worker

VERBOSITY_TO_LOG_LEVEL = {
    0: logging.CRITICAL,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


def reset_db_connections():
    for c in connections.all():
        c.close()


class Command(BaseCommand):
    """
    Runs RQ workers on specified queues. Note that all queues passed into a
    single rqworker command must share the same connection.

    Example usage:
    python manage.py rqworker high medium low
    """

    args = "<queue queue ...>"

    def add_arguments(self, parser):
        parser.add_argument(
            "--pid", action="store", dest="pidfile", default=None, help="file to write the worker`s pid into"
        )
        parser.add_argument(
            "--burst", action="store_true", dest="burst", default=False, help="Run worker in burst mode"
        )
        parser.add_argument("--name", action="store", dest="name", default=None, help="Name of the worker")
        parser.add_argument(
            "--worker-ttl",
            action="store",
            type=int,
            dest="worker_ttl",
            default=420,
            help="Default worker timeout to be used",
        )
        parser.add_argument(
            "--max-jobs",
            action="store",
            default=None,
            dest="max_jobs",
            type=int,
            help="Maximum number of jobs to execute before terminating worker",
        )
        parser.add_argument(
            "--fork-job-execution",
            action="store",
            default=True,
            dest="fork_job_execution",
            type=bool,
            help="Fork job execution to another process",
        )
        parser.add_argument("--job-class", action="store", dest="job_class", help="Jobs class to use")
        parser.add_argument(
            "queues",
            nargs="*",
            type=str,
            help="The queues to work on, separated by space, all queues should be using the same redis",
        )
        parser.add_argument("--sentry-dsn", action="store", dest="sentry_dsn", help="Sentry DSN to use")
        parser.add_argument("--sentry-debug", action="store_true", dest="sentry_debug", help="Enable Sentry debug mode")
        parser.add_argument("--sentry-ca-certs", action="store", dest="sentry_ca_certs", help="Path to CA certs file")

    def handle(self, **options):
        queues = options.get("queues", [])
        if not queues:
            queues = [
                "default",
            ]
        click.echo(f"Starting worker for queues {queues}")
        pidfile = options.get("pidfile")
        if pidfile:
            with open(os.path.expanduser(pidfile), "w") as fp:
                fp.write(str(os.getpid()))

        # Verbosity is defined by default in BaseCommand for all commands
        verbosity = options.get("verbosity", 1)
        log_level = VERBOSITY_TO_LOG_LEVEL.get(verbosity, logging.INFO)
        setup_loghandlers(log_level)

        try:
            # Instantiate a worker
            w = create_worker(
                *queues,
                name=options["name"],
                job_class=options.get("job_class"),
                default_worker_ttl=options["worker_ttl"],
                fork_job_execution=options["fork_job_execution"],
            )

            # Close any opened DB connection before any fork
            reset_db_connections()

            # Check whether sentry is enabled
            if options.get("sentry_dsn") is not None:
                sentry_opts = dict(ca_certs=options.get("sentry_ca_certs"), debug=options.get("sentry_debug"))
                register_sentry(options.get("sentry_dsn"), **sentry_opts)

            w.work(
                burst=options.get("burst", False),
                logging_level=log_level,
                max_jobs=options["max_jobs"],
            )
        except (redis.ConnectionError, valkey.ConnectionError) as e:
            click.echo(str(e), err=True)
            sys.exit(1)



from typing import Union, Dict, Tuple, Type

import redis
import valkey

from scheduler.settings import Broker

ConnectionErrorType = Union[redis.ConnectionError, valkey.ConnectionError]
ConnectionType = Union[redis.Redis, valkey.Valkey]
PipelineType = Union[redis.client.Pipeline, valkey.client.Pipeline]
RedisSentinel = redis.sentinel.Sentinel

BrokerConnectionClass: Dict[Tuple[Broker, bool], Type] = {
    # Map of (Broker, Strict flag) => Connection Class
    (Broker.REDIS, False): redis.Redis,
    (Broker.VALKEY, False): valkey.Valkey,
    (Broker.REDIS, True): redis.StrictRedis,
    (Broker.VALKEY, True): valkey.StrictValkey,
}



import logging
from dataclasses import dataclass
from enum import Enum
from typing import Callable

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

logger = logging.getLogger(__package__)

QUEUES = dict()


class Broker(Enum):
    REDIS = "redis"
    FAKEREDIS = "fakeredis"
    VALKEY = "valkey"


@dataclass
class SchedulerConfig:
    EXECUTIONS_IN_PAGE: int
    DEFAULT_RESULT_TTL: int
    DEFAULT_TIMEOUT: int
    SCHEDULER_INTERVAL: int
    BROKER: Broker
    TOKEN_VALIDATION_METHOD: Callable[[str], bool]


def _token_validation(token: str) -> bool:
    return False


SCHEDULER_CONFIG: SchedulerConfig = SchedulerConfig(
    EXECUTIONS_IN_PAGE=20,
    DEFAULT_RESULT_TTL=600,
    DEFAULT_TIMEOUT=300,
    SCHEDULER_INTERVAL=10,
    BROKER=Broker.REDIS,
    TOKEN_VALIDATION_METHOD=_token_validation,
)


def conf_settings():
    global QUEUES
    global SCHEDULER_CONFIG

    QUEUES = getattr(settings, "SCHEDULER_QUEUES", None)
    if QUEUES is None:
        logger.warning("Configuration using RQ_QUEUES is deprecated. Use SCHEDULER_QUEUES instead")
        QUEUES = getattr(settings, "RQ_QUEUES", None)
    if QUEUES is None:
        raise ImproperlyConfigured("You have to define SCHEDULER_QUEUES in settings.py")

    user_settings = getattr(settings, "SCHEDULER_CONFIG", {})
    if "FAKEREDIS" in user_settings:
        user_settings["BROKER"] = Broker.FAKEREDIS if user_settings["FAKEREDIS"] else Broker.REDIS
        user_settings.pop("FAKEREDIS")
    for k in user_settings:
        if k not in SCHEDULER_CONFIG.__annotations__:
            raise ImproperlyConfigured(f"Unknown setting {k} in SCHEDULER_CONFIG")
        setattr(SCHEDULER_CONFIG, k, user_settings[k])


conf_settings()



# Generated by Django 4.2.5 on 2023-09-08 20:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("scheduler", "0013_alter_cronjob_queue_alter_repeatablejob_queue_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="cronjob",
            name="created",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name="cronjob",
            name="modified",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="repeatablejob",
            name="created",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name="repeatablejob",
            name="modified",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="scheduledjob",
            name="created",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name="scheduledjob",
            name="modified",
            field=models.DateTimeField(auto_now=True),
        ),
    ]



from django.urls import reverse

from scheduler.tests.testtools import SchedulerBaseCase


class TestWorkerAdmin(SchedulerBaseCase):
    def test_admin_list_view(self):
        # arrange
        self.client.login(username="admin", password="admin")
        model = "worker"
        url = reverse(f"admin:scheduler_{model}_changelist")

        # act
        res = self.client.get(url)
        # assert
        self.assertEqual(200, res.status_code)


class TestQueueAdmin(SchedulerBaseCase):
    def test_admin_list_view(self):
        # arrange
        self.client.login(username="admin", password="admin")
        model = "queue"
        url = reverse(f"admin:scheduler_{model}_changelist")

        # act
        res = self.client.get(url)
        # assert
        self.assertEqual(200, res.status_code)



import sys

import click
from django.apps import apps
from django.core.management.base import BaseCommand

from scheduler.tools import MODEL_NAMES


class Command(BaseCommand):
    """
    Export all scheduled jobs
    """

    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "-o",
            "--output",
            action="store",
            choices=["json", "yaml"],
            default="json",
            dest="format",
            help="format of output",
        )

        parser.add_argument(
            "-e",
            "--enabled",
            action="store_true",
            dest="enabled",
            help="Export only enabled jobs",
        )
        parser.add_argument(
            "-f",
            "--filename",
            action="store",
            dest="filename",
            help="File name to load (otherwise writes to standard output)",
        )

    def handle(self, *args, **options):
        file = open(options.get("filename"), "w") if options.get("filename") else sys.stdout
        res = list()
        for model_name in MODEL_NAMES:
            model = apps.get_model(app_label="scheduler", model_name=model_name)
            jobs = model.objects.all()
            if options.get("enabled"):
                jobs = jobs.filter(enabled=True)
            for job in jobs:
                res.append(job.to_dict())

        if options.get("format") == "json":
            import json

            click.echo(json.dumps(res, indent=2), file=file)
            return

        if options.get("format") == "yaml":
            try:
                import yaml
            except ImportError:
                click.echo("Aborting. LibYAML is not installed.")
                exit(1)
            # Disable YAML alias
            yaml.Dumper.ignore_aliases = lambda *x: True
            click.echo(yaml.dump(res, default_flow_style=False), file=file)
            return



from django.core.exceptions import ValidationError

from scheduler import settings
from scheduler.models import CronTask
from scheduler.tools import create_worker
from .test_models import BaseTestCases
from .testtools import task_factory
from ..queues import get_queue


class TestCronTask(BaseTestCases.TestBaseTask):
    TaskModelClass = CronTask

    def test_clean(self):
        task = task_factory(CronTask)
        task.cron_string = "* * * * *"
        task.queue = list(settings.QUEUES)[0]
        task.callable = "scheduler.tests.jobs.test_job"
        self.assertIsNone(task.clean())

    def test_clean_cron_string_invalid(self):
        task = task_factory(CronTask)
        task.cron_string = "not-a-cron-string"
        task.queue = list(settings.QUEUES)[0]
        task.callable = "scheduler.tests.jobs.test_job"
        with self.assertRaises(ValidationError):
            task.clean_cron_string()

    def test_check_rescheduled_after_execution(self):
        task = task_factory(
            CronTask,
        )
        queue = task.rqueue
        first_run_id = task.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 0)
        self.assertIsNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 1)
        self.assertIsNotNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_id, first_run_id)

    def test_check_rescheduled_after_failed_execution(self):
        task = task_factory(
            CronTask,
            callable_name="scheduler.tests.jobs.scheduler.tests.jobs.test_job",
        )
        queue = task.rqueue
        first_run_id = task.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 1)
        self.assertIsNotNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 0)
        self.assertIsNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_id, first_run_id)

    def test_cron_task_enqueuing_jobs(self):
        queue = get_queue()
        prev_queued = len(queue.scheduled_job_registry)
        prev_finished = len(queue.finished_job_registry)
        task = task_factory(CronTask, callable_name="scheduler.tests.jobs.enqueue_jobs")
        self.assertEqual(prev_queued + 1, len(queue.scheduled_job_registry))
        first_run_id = task.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        self.assertEqual(20, len(queue))
        self.assertEqual(prev_finished + 1, len(queue.finished_job_registry))
        worker = create_worker(
            "default",
            fork_job_execution=False,
        )
        worker.work(burst=True)
        self.assertEqual(prev_finished + 21, len(queue.finished_job_registry))
        worker.refresh()
        self.assertEqual(20, worker.successful_job_count)
        self.assertEqual(0, worker.failed_job_count)
